package com.pancik.ciernypetrzlen;

import com.badlogic.gdx.ApplicationListener;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputMultiplexer;
import com.badlogic.gdx.graphics.GL10;
import com.badlogic.gdx.graphics.OrthographicCamera;
import com.badlogic.gdx.math.Rectangle;
import com.pancik.ciernypetrzlen.engine.SoundData;
import com.pancik.ciernypetrzlen.screens.MainGameScreen;
import com.pancik.ciernypetrzlen.screens.PickModeScreen;
import com.pancik.ciernypetrzlen.screens.Screen;
import com.pancik.ciernypetrzlen.util.HorizontalAlignment;
import com.pancik.ciernypetrzlen.util.VerticalAlignment;
import com.pancik.ciernypetrzlen.util.RenderUtils;

public class MainApplicationListener implements ApplicationListener {
    public class Controls {
        public void setScreen(Screen screen, boolean dispose) {
            doDispose = dispose;
            
            screen.create();
            lastScreen = currentScreen;
            currentScreen = screen;
            
            phasing = true;
            phasingTimeElapsed = 0.0f;

            Gdx.input.setOnscreenKeyboardVisible(false);
        }

        public OrthographicCamera getUICamera() {
            return uiCamera;
        }
        
        public InputMultiplexer getInputMultiplexer() {
            return inputMultiplexer;
        }
        
        public void playMenuMusic() {
            if(SoundData.SOUND_IS_ON) {
                SoundData.playMenuMusic();
            } else {
                SoundData.stopMenuMusic();
            }
        }
        
        public void playGameMusic() {
            if(SoundData.SOUND_IS_ON) {
                SoundData.playEngineMusic();
            } else {
                SoundData.stopEngineMusic();
            }
        }
    }
    
    private static final float PHASING_TIME = 0.250f;

    private boolean phasing = false;
    private float phasingTimeElapsed = 0.0f;
    
    private static Screen currentScreen = null;
    private static Screen lastScreen = null;
    private static boolean doDispose;
    private OrthographicCamera uiCamera;
    private Rectangle glViewport;
    private InputMultiplexer inputMultiplexer;

    private float accum = 0.0f;
    
    public MainApplicationListener() {
    }
    
    @Override
    public void create() {
        this.inputMultiplexer = new InputMultiplexer();
        Gdx.input.setInputProcessor(inputMultiplexer);

        Gdx.input.setCatchBackKey(true);
        Gdx.input.setCatchMenuKey(true);

        DrawableData.loadData();
        SoundData.loadData();
        
        glViewport = new Rectangle(0, 0, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());

        uiCamera = new OrthographicCamera(Gdx.graphics.getWidth(), Gdx.graphics.getHeight());
        uiCamera.setToOrtho(true, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());
        uiCamera.position.set(Gdx.graphics.getWidth()/2, Gdx.graphics.getHeight()/2, 0);

        Controls controls = new Controls();
        if(currentScreen == null) {
            Screen setScreen = new PickModeScreen(controls);
            controls.setScreen(setScreen, true);
            phasingTimeElapsed = PHASING_TIME/2;
        } else {
            currentScreen.updateMainControls(controls);
        }
    }

    private static final float STEP_TIME = 1.0f/60.0f;
    
    @Override
    public void render() {
        //long startTimeTotal = TimeUtils.nanoTime();
        accum += Gdx.graphics.getDeltaTime();
        
        if(phasing) {
            if(phasingTimeElapsed + Gdx.graphics.getDeltaTime() >= PHASING_TIME) {
                phasing = false;
            } else {
                phasingTimeElapsed += Gdx.graphics.getDeltaTime();
            }
        }
        
        while (accum >= STEP_TIME) {
            currentScreen.tick();
            accum -= STEP_TIME;
        }
        
        boolean useLast = false;
        if(phasing && lastScreen != null) {
            useLast = (phasingTimeElapsed < PHASING_TIME/2) ? true : false;
        }
        
        //before rendering update camera
        Gdx.gl.glClear(GL10.GL_COLOR_BUFFER_BIT | GL10.GL_DEPTH_BUFFER_BIT);
        if(!useLast) {
            if(!(currentScreen instanceof MainGameScreen)) {
                Gdx.gl.glClearColor(0.6274f, 0.6f, 0.2862f, 1.0f);
            } else {
                Gdx.gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
            }
        } else {
            if(lastScreen != null && !(lastScreen instanceof MainGameScreen)) {
                Gdx.gl.glClearColor(0.6274f, 0.6f, 0.2862f, 1.0f);
            } else {
                Gdx.gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
            }
        }
        Gdx.gl.glViewport((int)this.glViewport.x,
                (int)this.glViewport.y,
                (int)this.glViewport.width,
                (int)this.glViewport.height);
        
        //render screen
        if(!useLast) {
            currentScreen.render();
            
            if(lastScreen != null) {
                if(doDispose) {
                    lastScreen.dispose();
                }
                lastScreen = null;
            }
        } else {
            lastScreen.render();
        }
        
        this.uiCamera.update();
        DrawableData.spriteBatch.setProjectionMatrix(uiCamera.combined);
        DrawableData.shapeRenderer.setProjectionMatrix(uiCamera.combined);
        DrawableData.spriteBatch.begin();
        
        if(!useLast) {
            currentScreen.renderUI();
        } else {
            lastScreen.renderUI();
        }
        
        /*
        DrawableData.font.setScale(3.0f);
        RenderUtils.blitText("fps: " + Gdx.graphics.getFramesPerSecond(),
                HorizontalAlignment.RIGHT, -2,
                VerticalAlignment.TOP, 2,
                RenderUtils.TextAlligment.RIGHT,
                RenderUtils.TextAlligment.BOTTOM);
        */
        
        if(phasing) {
            boolean darkening = (phasingTimeElapsed < PHASING_TIME/2) ? true : false;
            float step = darkening ? phasingTimeElapsed : phasingTimeElapsed - PHASING_TIME/2;
            float max = PHASING_TIME/2;
            
            if(darkening) {
                DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, ((1.0f/max)*step));
            } else {
                DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 1.0f - (1.0f/max)*step);
            }
            
            RenderUtils.blit(DrawableData.textureAtlas.findRegion("solid-black"),
                    0, 0, Gdx.graphics.getWidth(), Gdx.graphics.getHeight(), 0.0f, false);
            
            DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 1.0f);
        }
        
        DrawableData.spriteBatch.end();
        //System.out.println("Update: " + (TimeUtils.nanoTime() - startTimeTotal) / 1000000000.0f);
    }

    @Override
    public void pause() {
        currentScreen.pause();
    }

    @Override
    public void resume() {
        currentScreen.resume();
    }

    @Override
    public void dispose() {
        currentScreen.dispose();
        
        DrawableData.dispose();
        SoundData.dispose();
    }

    @Override
    public void resize(int width, int heigth) {
        uiCamera = new OrthographicCamera();
        uiCamera.setToOrtho(true, width, heigth);
        uiCamera.position.set(width/2, heigth/2, 0);
        
        glViewport = new Rectangle(0, 0, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());

        currentScreen.resize(width, heigth);
    }

    public InputMultiplexer getInputMultiplexer() {
        return inputMultiplexer;
    }
}