package com.pancik.ciernypetrzlen.engine;

import com.pancik.ciernypetrzlen.engine.component.level.Level;
import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.GL10;
import com.badlogic.gdx.graphics.PerspectiveCamera;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.g3d.decals.CameraGroupStrategy;
import com.badlogic.gdx.graphics.g3d.decals.Decal;
import com.badlogic.gdx.graphics.g3d.decals.DecalBatch;
import com.badlogic.gdx.math.Intersector;
import com.badlogic.gdx.math.Plane;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.math.collision.Ray;
import com.badlogic.gdx.utils.Disposable;
import com.pancik.ciernypetrzlen.DrawableData;
import com.pancik.ciernypetrzlen.MainApplicationListener;
import com.pancik.ciernypetrzlen.engine.component.EntityManager;
import com.pancik.ciernypetrzlen.engine.component.LevelManager;
import com.pancik.ciernypetrzlen.engine.component.entity.Entity;
import com.pancik.ciernypetrzlen.engine.pathfinding.CollisionMap;
import com.pancik.ciernypetrzlen.engine.player.Player;
import com.pancik.ciernypetrzlen.screens.MainGameScreen;
import com.pancik.ciernypetrzlen.screens.MainMenuScreen;
import java.util.Comparator;

public class Engine implements Disposable {
    public class Controls {
        public LevelManager getLevelManager() {
            return levelManager;
        }
        
        public EntityManager getEntityManager() {
            return entityManager;
        }

        public void addEntity(Entity e) {
            getEntityManager().addEntity(e);
        }

        public Player getPlayer() {
            return player;
        }
        
        public void pauseGame(boolean pause) {
            gamePaused = pause;
        }
        
        public CollisionMap getCollisionMap() {
            return collisionMap;
        }
        
        public Level getLevel() {
            return levelData;
        }
        
        public DecalBatch getDecalBatch() {
            return decalBatch;
        }
        
        public Decal createDecal(Vector2 position, float sizeX, float sizeY, TextureRegion region) {
            Decal decal = Decal.newDecal(sizeX, sizeY, region, true);
            decal.setPosition(position.x, 0.20f, position.y);
            decal.lookAt(decal.getPosition().cpy().add(camera.direction), Vector3.Y);
            return decal;
        }
        
        public void startLevelUpEffect() {
            SoundData.playSound("level-up", 0.5f);
            levelupEffect = true;
        }
        
        public Color getLightColor() {
            return lightColor;
        }
        
        public float getLightIntensity() {
            return lightIntensity;
        }
        
        public PerspectiveCamera getCamera() {
            return camera;
        }
     
        public void focusCamera(Vector2 position) {
            ray.set(camera.position, camera.direction);
            Intersector.intersectRayPlane(ray, plane, intersection);
            camera.position.sub(intersection.sub(position.x, 0.0f, position.y));
        }

        public Vector3 getCameraLookAt() {
            ray.set(camera.position, camera.direction);
            Intersector.intersectRayPlane(ray, plane, lookAt);
            return lookAt;
        }
        
        public void reload() {
            Engine.this.reload();
        }
        
        public MainApplicationListener.Controls getMainControls() {
            return mainControls;
        }
        
        public void goToMainMenu() {
            mainControls.setScreen(new MainMenuScreen(mainControls), true);
        }
    }
    
    private MainApplicationListener.Controls mainControls;
    private boolean reload = false;
    
    private Level levelData;
    private LevelManager levelManager;
    private EntityManager entityManager;
    private Player player;
    private CollisionMap collisionMap;
    private boolean gamePaused = false;
    
    private PerspectiveCamera camera;
    private final DecalBatch decalBatch;
    
    private Color lightColor;
    private float lightIntensity = 0.04f;
    private boolean levelupEffect = false;
    private boolean levelupLowering = true;
    private final float normalLightIntensity = 0.04f;
    private final float levelupIntensity = 0.005f;
    
    private final Vector3 intersection = new Vector3();
    private final Plane plane = new Plane(new Vector3(0.0f, 1.0f, 0.0f), 0.0f);
    private Ray ray = new Ray(new Vector3(), new Vector3());
    private Vector3 lookAt = new Vector3();
    
    private PersistentData persistentData;
    
    public Engine(MainApplicationListener.Controls mainControls, PersistentData data) {
        this.mainControls = mainControls;
        this.persistentData = data;
        
        this.camera = new PerspectiveCamera(67.0f, Gdx.graphics.getWidth(), Gdx.graphics.getHeight());
        this.camera.near = 1.0f;
        this.camera.far = 100.0f;
        this.camera.position.set(0.0f, 6.666667f, 2.0f);
        this.camera.lookAt(0.0f, 0.0f, 0.0f);
        
        this.decalBatch = new DecalBatch(new CameraGroupStrategy(this.camera, new Comparator<Decal>() {
            @Override
            public int compare(Decal d1, Decal d2) {
                return (int) Math.signum((d1.getPosition().z) - (d2.getPosition().z));
            }
        }));

        this.lightColor = new Color(1.0f, 0.8f, 0.7f, 1.0f);
    }
    
    public void create() {
        levelData = new Level(Gdx.files.internal("data/level/" + persistentData.currentLevel), new Controls());
        collisionMap = levelData.buildCollisionMap();
        
        levelManager = new LevelManager(levelData, new Controls());
        entityManager = new EntityManager(levelData, new Controls());
        player = new Player(new Controls(), persistentData);
    }
    
    public void tick() {
        if(!gamePaused) {
            levelManager.tick();

            entityManager.tick();

            player.tick();
        }
        
        player.tickUI();
        
        if(reload) {
            mainControls.setScreen(new MainGameScreen(mainControls, PersistentData.load()), true);
        }
    }
    
    public void render() {
        if(levelupEffect) {
            float delta = (normalLightIntensity - levelupIntensity)/40.0f;
            if(levelupLowering) {
                lightIntensity -= delta * 3.0f;
                if(lightIntensity <= levelupIntensity) {
                    levelupLowering = false;
                }
            } else {
                lightIntensity += delta;
                if(lightIntensity >= normalLightIntensity) {
                    lightIntensity = normalLightIntensity;
                    levelupLowering = true;
                    levelupEffect = false;
                    player.yellLevelUp();
                }
            }
        }
        
        camera.update();
        DrawableData.shapeRenderer.setProjectionMatrix(camera.combined);
        DrawableData.shapeRenderer.rotate(1.0f, 0.0f, 0.0f, 90.0f);
        Gdx.gl.glEnable(GL10.GL_DEPTH_TEST);
        
        levelManager.render();
        
        entityManager.render();

        Gdx.gl.glDisable(GL10.GL_DEPTH_TEST);
        DrawableData.shapeRenderer.identity();
    }
    
    public void renderUI() {
        player.renderUI();
    }
    
    @Override
    public void dispose() {
        levelManager.dispose();
        
        entityManager.dispose();
        
        player.dispose();
        
        decalBatch.dispose();
    }
    
    public void resize(int sizeX, int sizeY) {
        camera.viewportWidth = sizeX;
        camera.viewportHeight = sizeY;
        
        player.resize(sizeX, sizeY);
    }
    
    public void reload() {
        this.reload = true;
    }
}