package com.pancik.ciernypetrzlen.engine.component;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.GL10;
import com.badlogic.gdx.graphics.g3d.decals.Decal;
import com.badlogic.gdx.graphics.g3d.decals.DecalBatch;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.math.Vector3;
import com.badlogic.gdx.utils.Disposable;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.component.entity.Attackable;
import com.pancik.ciernypetrzlen.engine.component.entity.Entity;
import com.pancik.ciernypetrzlen.engine.component.entity.Interactable;
import com.pancik.ciernypetrzlen.engine.component.entity.Unit;
import com.pancik.ciernypetrzlen.engine.component.level.Level;
import java.util.ArrayList;
import java.util.List;

public class EntityManager implements Disposable {
    private Engine.Controls engineControls;
    
    private final List<Entity> entities = new ArrayList<Entity>();

    private boolean renderDebug = false;
    private InputHandler inputHandler = new InputHandler();
    
    public EntityManager(Level level, Engine.Controls engineControls) {
        this.engineControls = engineControls;
        
        this.engineControls.getMainControls().getInputMultiplexer().addProcessor(0, inputHandler);
        
        for(Entity e : level.getEntities()) {
            addEntity(e);
        }
    }
    
    public void tick() {
        for(int i = 0; i < entities.size(); i++) {
            if(!entities.get(i).remove()) {
                entities.get(i).tick();
            }
        }
        
        ArrayList<Entity> shouldBeRemoved = new ArrayList<Entity>();

        for(Entity e : entities) {
            if(e.remove()) {
                shouldBeRemoved.add(e);
            }
        }

        for(Entity e : shouldBeRemoved) {
            removeEntity(e);
        }
    }
    
    public void render() {
        Vector3 lookAt = engineControls.getCameraLookAt().add(0.0f, 4.0f, 0.0f);
        DecalBatch decalBatch = engineControls.getDecalBatch();
        for(Entity e : entities) {
            if(!e.remove()) {
                boolean highlighted = false;
                if(e instanceof Interactable) {
                    highlighted = ((Interactable) e).isHighlighted();
                }
                
                Decal decal = e.getDecal();
                if(decal != null) {
                    float dist = lookAt.dst(decal.getPosition());
                    float i = 1.0f * (1.0f / ((engineControls.getLightIntensity() + 0.01f) * dist * dist));
                    i = i > 1.0f ? 1.0f : i;
                    if(!highlighted) {
                        decal.setColor(i, i, i, 1.0f);
                    } else {
                        decal.setColor(1.0f, 0.5f, 0.5f, 1.0f);
                    }
                    decalBatch.add(decal);
                }
                
                e.render();
            }
        }
        decalBatch.flush();
        
        if(renderDebug) {
            Gdx.gl.glDisable(GL10.GL_DEPTH_TEST);
            for(Entity e : entities) {
                e.debugRender();
            }
            Gdx.gl.glEnable(GL10.GL_DEPTH_TEST);
        }
    }
    
    public void addEntity(Entity e) {
        e.create();
        entities.add(e);
    }
    
    public void removeEntity(Entity e) {
        e.onRemove();
        entities.remove(e);
        if(e instanceof Disposable) {
            ((Disposable) e).dispose();
        }
    }
    
    public Entity findById(Long id) {
        Entity out = null;
        
        for(Entity entity : entities) {
            if(entity.getId().equals(id)) {
                out = entity;
                break;
            }
        }
        
        return out;
    }
        
    public List<Attackable> getAttackableEntities(Vector2 from, float range) {
        List<Attackable> out = new ArrayList<Attackable>();
        
        for(Entity e : entities) {
            if(e instanceof Attackable) {
                Attackable attackableEntity = (Attackable)e;
                
                if(attackableEntity.getPosition().dst(from) < range) {
                    out.add(attackableEntity);
                }
            }
        }
        
        return out;
    }
 
    public List<Unit> getUnits(Vector2 from, float range) {
        List<Unit> out = new ArrayList<Unit>();

        for (Entity e : entities) {
            if (e instanceof Unit) {
                Unit unit = (Unit) e;

                if (unit.getPosition().dst(from) < range) {
                    out.add(unit);
                }
            }
        }

        return out;
    }

    public List<Entity> getEntities() {
        return entities;
    }
    
    public int getEntityCount() {
        return entities.size();
    }
    
    @Override
    public void dispose() {
        this.engineControls.getMainControls().getInputMultiplexer().removeProcessor(inputHandler);
        
        for(Entity e : entities) {
            if(e instanceof Disposable) {
                ((Disposable) e).dispose();
            }
        }
    }
    
    private class InputHandler extends InputAdapter {
        @Override
        public boolean keyDown(int keycode) {
            if(keycode == Input.Keys.P) {
                renderDebug = !renderDebug;
            }
            return false;
        }
    }
}
