package com.pancik.ciernypetrzlen.engine.component.entity;

import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.graphics.g3d.decals.Decal;
import com.badlogic.gdx.graphics.glutils.ShapeRenderer;
import com.badlogic.gdx.math.Vector2;
import com.pancik.ciernypetrzlen.DrawableData;
import com.pancik.ciernypetrzlen.engine.Animation;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.component.entity.behaviour.MonsterBehaviour;
import com.pancik.ciernypetrzlen.engine.component.entity.loot.Loot;
import com.pancik.ciernypetrzlen.engine.component.entity.loot.MonsterLoot;
import com.pancik.ciernypetrzlen.engine.component.entity.projectile.Projectile;
import com.pancik.ciernypetrzlen.engine.component.spell.Buff;
import com.pancik.ciernypetrzlen.engine.pathfinding.BresenhamCollisionChecker;

public abstract class Monster extends Unit implements Interactable {
    public static class MonsterStats {
        public int health;
        public float speed;
        public int attackDamage;
        public float attackRange;
        public int attackStep;
        public int attackCooldownLength;
        public int level;
        public int experience;

        public MonsterStats(int level, int health, float experienceCoef, float speed, int attackDamage, float attackRange, int attackStep, int attackCooldownLength) {
            this.level = level;
            this.health = health + (int)(((float)health/5.0f) * level);
            this.attackDamage = attackDamage + Math.round((float)Math.pow((float)level, 1.4f));
            this.experience = (int)((float)level * experienceCoef);
            
            this.speed = speed;
            this.attackRange = attackRange;
            this.attackStep = attackStep;
            this.attackCooldownLength = attackCooldownLength;
        }
    }
    
    protected MonsterStats stats;
    protected int attackCooldown = 0;
    
    public Monster(Vector2 position, Vector2 size, int team, Engine.Controls engineControls, MonsterStats stats) {
        super(position, size, team, stats.health, stats.speed, new MonsterLoot(stats.level), new MonsterBehaviour(), engineControls);
        this.stats = stats;
    }
    
    public Monster(Vector2 position, Vector2 size, int team, Loot loot, Engine.Controls engineControls, MonsterStats stats) {
        super(position, size, team, stats.health, stats.speed, loot, new MonsterBehaviour(), engineControls);
        this.stats = stats;
    }
    
    @Override
    public void create() {
        decal = engineControls.createDecal(position, size.x, size.y, getStandAnimation()[Animation.DOWN][0]);
    }

    @Override
    public void tick() {
        super.tick();
        
        this.speed = (stats.speed + getScalarModifier(Buff.Stat.MOVEMENT_SPEED)) * getPercentualModifier(Buff.Stat.MOVEMENT_SPEED);
        
        if(attackCooldown > 0) {
            attackCooldown--;
        }
    }

    @Override
    public Decal getDecal() {
        TextureRegion[][] textures;
        int step;
        if(currentAction == Action.WALK || currentAction == Action.STAND) {
            textures = currentAction == Action.WALK ? getWalkAnimation() : getStandAnimation();
            step = animationStep / 20;
        } else {
            textures = getAttackAnimation();
            step = animationStep / stats.attackStep;
        }
        decal.setTextureRegion(Animation.getTexture(textures, direction, step));
        decal.setPosition(position.x, decal.getY(), position.y);
        return decal;
    }

    @Override
    public void debugRender() {
        DrawableData.shapeRenderer.begin(ShapeRenderer.ShapeType.Line);
        DrawableData.shapeRenderer.setColor(1.0f, 0.0f, 0.0f, 1.0f);
        DrawableData.shapeRenderer.rect(position.x - size.x/2, position.y - size.y/2, size.x, size.y);
        DrawableData.shapeRenderer.line(position.x, position.y, position.x + direction.x, position.y + direction.y);
        DrawableData.shapeRenderer.setColor(1.0f, 1.0f, 0.0f, 1.0f);
        DrawableData.shapeRenderer.line(position.x, position.y, position.x + (velocity.x * 10.0f), position.y + (velocity.y * 10.0f));
        
        if(path != null) {
            DrawableData.shapeRenderer.setColor(1.0f, 0.0f, 1.0f, 1.0f);
            for(Vector2 v : path) {
                DrawableData.shapeRenderer.circle(v.x, v.y, 0.25f, 10);
            }
        }
        
        DrawableData.shapeRenderer.end();
    }

    @Override
    public boolean remove() {
        return isDead();
    }

    @Override
    public boolean isInRangeForAttack(Attackable a) {
        return (position.dst(a.getPosition()) < stats.attackRange);
    }

    @Override
    public void interact() {
    }
    
    @Override
    public void attack(Attackable a) {
        direction.set(a.getPosition()).sub(position).nor();
    }
    
    @Override
    protected void doAttack(int damage, Attackable a) {
        if (actionListener != null) {
            actionListener.damageDealtAttack(damage, a);
        }
    }
    
    @Override
    public int getHit(int damage, Attackable from) {
        int reducedDamage = Math.round((float)damage * (1.0f - (getScalarModifier(Buff.Stat.DEFENSE)) * getPercentualModifier(Buff.Stat.DEFENSE)));
        return super.getHit(reducedDamage, from);
    }

    @Override
    public int getRewardExperience() {
        return stats.experience;
    }
    
    @Override
    public String getName() {
        return "(" + stats.level + ")";
    }
    
    protected abstract TextureRegion[][] getStandAnimation();
    protected abstract TextureRegion[][] getWalkAnimation();
    protected abstract TextureRegion[][] getAttackAnimation();
    
    public static abstract class Melee extends Monster {
        public Melee(Vector2 position, Vector2 size, int team, Engine.Controls engineControls, MonsterStats stats) {
            super(position, size, team, engineControls, stats);
        }

        @Override
        public void attack(Attackable a) {
            super.attack(a);
            
            if (attackCooldown == 0) {
                setAttack(getAttackAnimation()[0].length * stats.attackStep, 3 * stats.attackStep, a,
                        (int)(((float)stats.attackDamage + getScalarModifier(Buff.Stat.ATTACK)) * getPercentualModifier(Buff.Stat.ATTACK)));
                attackCooldown = (int)((float)stats.attackCooldownLength * (1.0f/((1.0f + getScalarModifier(Buff.Stat.ATTACK_SPEED)) * getPercentualModifier(Buff.Stat.ATTACK_SPEED))));
            }
        }

        @Override
        protected void doAttack(int damage, Attackable a) {
            int dealtDamage = a.getHit(damage, a);
            super.doAttack(dealtDamage, a);
        }
    }
    
    public static abstract class Ranged extends Monster {
        protected int failedShots = 0;
        protected int failedShotsRecoveryStep = 0;
        protected final int maxFailedShots = 2;
        protected final int failedShotsRecoveryLength = 180;
        
        public Ranged(Vector2 position, Vector2 size, int team, Engine.Controls engineControls, MonsterStats stats) {
            super(position, size, team, engineControls, stats);
        }
    
        @Override
        public void tick() {
            super.tick();

            if (failedShotsRecoveryStep > 0) {
                failedShotsRecoveryStep--;
            }
        }

        @Override
        public boolean isInRangeForAttack(Attackable a) {
            return super.isInRangeForAttack(a)
                    && BresenhamCollisionChecker.isInLineOfSight(engineControls.getCollisionMap(), position.x, position.y, a.getPosition().x, a.getPosition().y)
                    && failedShotsRecoveryStep == 0;
        }

        @Override
        public void attack(Attackable a) {
            super.attack(a);

            if (attackCooldown == 0) {
                setAttack(getAttackAnimation()[0].length * stats.attackStep, 3 * stats.attackStep, a,
                        (int)(((float)stats.attackDamage + getScalarModifier(Buff.Stat.ATTACK)) * getPercentualModifier(Buff.Stat.ATTACK)));
                attackCooldown = (int)((float)stats.attackCooldownLength * (1.0f/((1.0f + getScalarModifier(Buff.Stat.ATTACK_SPEED)) * getPercentualModifier(Buff.Stat.ATTACK_SPEED))));
            }
        }

        @Override
        protected void doAttack(int damage, Attackable a) {
            final Attackable target = a;
            final int targetDamage = damage;
            createProjectile(a, new Projectile.Callback() {
                @Override
                public void onProjectileArrived(Vector2 pos) {
                    int dealtDamage = target.getHit(targetDamage, target);
                    if (actionListener != null) {
                        actionListener.damageDealtAttack(dealtDamage, target);
                    }

                    failedShots = 0;
                }

                @Override
                public void onProjectileCrashed(Vector2 pos) {
                    failedShots++;
                    if (failedShots > maxFailedShots) {
                        failedShots = 0;
                        failedShotsRecoveryStep = failedShotsRecoveryLength;
                    }
                }
            });
        }
        
        protected abstract void createProjectile(Attackable target, Projectile.Callback callback);
    }
}