package com.pancik.ciernypetrzlen.engine.component.entity;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.Array;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.component.entity.behaviour.Behaviour;
import com.pancik.ciernypetrzlen.engine.component.entity.loot.Loot;

public abstract class Unit extends Attackable {
    public static enum State {
        IDLE,
        WALK,
        ATTACK;
    }
    
    public static enum Action {
        STAND,
        WALK,
        ATTACK
    }
    
    public static interface ActionListener {
        public void damageDealtAttack(int damage, Attackable attackable);
        public void damageDealtSpell(int damage, Attackable attackable);
        public void damageRecieved(int damage, Attackable attackable);
    }
    
    protected Behaviour behaviour;
    protected State state;
    protected int stateTime;
    protected Array<Vector2> path;
    protected final Vector2 acceleration;
    protected final Vector2 velocity;
    protected final Vector2 direction;
    protected float speed;
    protected ActionListener actionListener;
    
    protected Action currentAction;
    protected int animationStep = 0;
    protected int animationAttackSteps;
    protected int animationAttackDelay;
    protected Attackable animationAttackable = null;
    protected int animationAttackDamage;
    
    public Unit(Vector2 pos, Vector2 size, int team, int health, float speed, Loot loot, Behaviour behaviour, Engine.Controls engineControls) {
        super(pos, size, team, health, loot, engineControls);
        
        this.speed = speed;
        this.acceleration = new Vector2();
        this.velocity = new Vector2();
        this.direction = new Vector2(0.0f, 1.0f);
        this.behaviour = behaviour;
        this.state = State.IDLE;
        this.stateTime = 0;
    }

    @Override
    public void tick() {
        super.tick();
        
        stateTime++;
        
        behaviour.tick(engineControls, this);

        velocity.x += acceleration.x;
        velocity.y += acceleration.y;
        if (velocity.len2() > speed * speed) {
            velocity.nor().scl(speed);
        }

        if(!collides(position.x + velocity.x, position.y)) {
            position.x += velocity.x;
        } else {
            //velocity.set(0.0f, velocity.y);
        }
        
        if(!collides(position.x, position.y + velocity.y)) {
            position.y += velocity.y;
        } else {
            //velocity.set(velocity.x, 0.0f);
        }
        
        if (velocity.len2() != 0.0f) {
            //direction.set(this.velocity).nor();
            direction.lerp(velocity, 0.90f).nor();
        }
        
        acceleration.set(0.0f, 0.0f);
        
        if(this.state != State.WALK) {
            //acceleration.div(30.0f);
            velocity.div(30.0f);
        }
        
        if(currentAction != Action.ATTACK) {
            if(velocity.len() > 0.001f) {
                setAnimation(Action.WALK);
            } else {
                setAnimation(Action.STAND);
            }
        } else {
            if(animationStep == animationAttackDelay && animationAttackable != null) {
                doAttack(animationAttackDamage, animationAttackable);
            }
            
            if(animationStep >= animationAttackSteps) {
                setAnimation(Action.STAND);
            }
        }
        animationStep++;
    }
    
    public void followPath() {
        if (path == null || currentAction == Action.ATTACK) {
            return;
        }

        Vector2 desiredVelocity = tmp2.set((Vector2) path.get(0)).sub(position);

        if (desiredVelocity.len() < 0.1f) {
            path.removeIndex(0);
            //System.out.println(path.size + " way points left");
            if (path.size == 0) {
                setPath(null);
                if(this instanceof Hero) {
                    ((Hero) this).setTargetPosition(null);
                }
                return;
            }
            desiredVelocity.set((Vector2) path.get(0)).sub(position);
        }

        desiredVelocity.nor().scl(getSpeed()).sub(velocity);
        acceleration.add(desiredVelocity);
    }

    public void moveTo(float x, float y) {
        if(currentAction == Action.ATTACK) {
            return;
        }
        
        Vector2 desiredVelocity = tmp2.set(x, y).sub(position);
        desiredVelocity.nor().scl(getSpeed()).sub(velocity);
        acceleration.add(desiredVelocity);
    }
    
    public void setPath(Array<Vector2> path) {
        this.path = path;
    }

    public Array<Vector2> getPath() {
        return path;
    }

    public void setState(State state) {
        this.state = state;
    }

    public State getState() {
        return state;
    }

    public int getStateTime() {
        return stateTime;
    }

    public Vector2 getAcceleration() {
        return acceleration;
    }

    public Vector2 getVelocity() {
        return velocity;
    }

    public Vector2 getDirection() {
        return direction;
    }

    public float getSpeed() {
        return speed;
    }

    @Override
    public void onRemove() {
        super.onRemove();
        
        behaviour.dispose();
    }
    
    protected void setAttack(int animationSteps, int animationDelay, Attackable target, int damage) {
        if (currentAction != Action.ATTACK) {
            animationAttackSteps = animationSteps;
            animationAttackDelay = animationDelay;
            animationAttackable = target;
            animationAttackDamage = damage;
            setAnimation(Action.ATTACK);
        }
    }
    
    protected void setAnimation(Action animation) {
        if(currentAction != animation) {
            animationStep = 0;
        }
        
        currentAction = animation;
    }

    public void setActionListener(ActionListener actionListener) {
        this.actionListener = actionListener;
    }

    public ActionListener getActionListener() {
        return actionListener;
    }
    
    private boolean collides(float posx, float posy) {
        /*
        float sizex = size.x/4.0f;
        float sizey = size.y/4.0f;
        float[] dx = {sizex, -sizex, 0.0f, 0.0f, sizex, -sizex, sizex, -sizex};
        float[] dy = {0.0f, 0.0f, sizey, -sizey, sizey, sizey, -sizey, -sizey};
        
        CollisionMap map = engineControls.getCollisionMap();
        for(int i = 0; i < dx.length; i++) {
            if(!map.isWalkable(posx + dx[i], posy + dy[i])) {
                return true;
            }
        }
        return false;
        */
        return !engineControls.getCollisionMap().isWalkable(posx, posy);
    }

    @Override
    public int getHit(int damage, Attackable from) {
        int damageRecieved = super.getHit(damage, from);

        if(actionListener != null) {
            actionListener.damageRecieved(damageRecieved, from);
        }
        return damageRecieved;
    }
    
    public abstract boolean isInRangeForAttack(Attackable a);
    public abstract void attack(Attackable a);
    protected abstract  void doAttack(int damage, Attackable a);
}
