package com.pancik.ciernypetrzlen.engine.component.entity.behaviour;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.Pool;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.component.entity.Attackable;
import com.pancik.ciernypetrzlen.engine.component.entity.Unit;
import com.pancik.ciernypetrzlen.engine.pathfinding.AStarPathFinder;
import com.pancik.ciernypetrzlen.engine.pathfinding.Cell;
import com.pancik.ciernypetrzlen.engine.pathfinding.CollisionMap;
import com.pancik.ciernypetrzlen.engine.pathfinding.DistanceEstimator;
import java.util.LinkedList;

public class MonsterBehaviour implements Behaviour {
    static final Pool<Vector2> vectorPool = new Pool() {
        @Override
        protected Vector2 newObject() {
            return new Vector2();
        }

        @Override
        public Vector2 obtain() {
            Vector2 vector = (Vector2) super.obtain();
            vector.x = 0.0f;
            vector.y = 0.0f;
            return vector;
        }
    };
    
    private static final float REACT_DISTANCE = 10.0F;
    private static final Vector2 tmp = new Vector2();

    private Attackable target = null;
    private boolean doSeek = false;
    private int seekCheckSteps = 0;
    private boolean stall = false;
    private int stallCheckStep = 0;
    private int stallSteps = 0;
    private int stallsOccured = 0;
    private LinkedList<Vector2> visited = new LinkedList<Vector2>();
    
    @Override
    public void tick(Engine.Controls engineControls, Unit unit) {
        if(target != null) {
            if(target.getPosition().dst(unit.getPosition()) >= REACT_DISTANCE || target.isDead()) {
                target = null;
                unit.setPath(null);
                unit.setState(Unit.State.IDLE);
            }
        }
        
        if(unit.getState() == Unit.State.IDLE) {
            for(Attackable a : engineControls.getEntityManager().getAttackableEntities(unit.getPosition(), REACT_DISTANCE)) {
                if(!a.isDead() && a.getTeam() != Attackable.DEBRIS_TEAM && a.getTeam() != unit.getTeam()) {
                    target = a;
                    unit.setState(Unit.State.WALK);
                    break;
                }
            }
        } else if(unit.getState() == Unit.State.WALK) {
            float distance = target.getPosition().dst(unit.getPosition());
            if(distance > REACT_DISTANCE) {
                target = null;
                resetStall();
                unit.setPath(null);
                unit.setState(Unit.State.IDLE);
                return;
            }
            if(unit.isInRangeForAttack(target)) {
                resetStall();
                unit.setState(Unit.State.ATTACK);
                return;
            }
            
            seekCheckSteps++;
            if(seekCheckSteps > 120) {
                doSeek = false;
                getPath(engineControls, unit);
                if(unit.getPath() != null) {
                    doSeek = true;
                    unit.setPath(null);
                }
                seekCheckSteps = 0;
            }
            
            if(doSeek) {
                checkStall(unit);
                if(!stall) {
                    seek(engineControls, unit);
                    separate(engineControls, unit);
                } else {
                    getPath(engineControls, unit);
                    followPath(unit.getPath(), unit);

                    stallSteps++;
                    if(stallSteps > 60 * (stallsOccured + 1)) {
                        stall = false;
                        unit.setPath(null);
                    }
                }
            }
        } else if(unit.getState() == Unit.State.ATTACK) {
            if(!unit.isInRangeForAttack(target)) {
                unit.setPath(null);
                unit.setState(Unit.State.IDLE);
            } else {
                unit.attack(target);
            }
        }
    }
    
    private void resetStall() {
        stall = false;
        stallsOccured = 0;
    }
    
    private void checkStall(Unit u) {
        stallCheckStep++;
        if(stallCheckStep > 30) {
            stallCheckStep = 0;
            final int maxOccurences = 2;
            final int maxVisitiedSize = 10;
            int occurences = 0;
            float x = (int)u.getPosition().x;
            float y = (int)u.getPosition().y;

            for(Vector2 v : visited) {
                if(v.x == x && v.y == y) {
                    occurences++;
                }
            }

            if(occurences > maxOccurences) {
                if(!stall) {
                    stallsOccured++;
                }
                stall = true;
                stallSteps = 0;
            }

            Vector2 vec = vectorPool.obtain();
            vec.set(x, y);
            visited.add(vec);
            if(visited.size() > maxVisitiedSize) {
                vectorPool.free(visited.getFirst());
                visited.removeFirst();
            }
        }
    }

    private Array<Vector2> getPath(Engine.Controls engineControls, Unit u) {
        Array path = u.getPath();

        boolean recalculate = (path == null) || (((Vector2) path.get(path.size - 1)).dst2(target.getPosition()) > 4.0f);
        if (recalculate) {
            CollisionMap map = engineControls.getCollisionMap();
            Cell startCell = map.getCell(u.getPosition().x, u.getPosition().y);
            Cell endCell = map.getCell(target.getPosition().x, target.getPosition().y);
            path = AStarPathFinder.findPath(startCell, endCell, DistanceEstimator.EUCLIDIAN, DistanceEstimator.MANHATTEN);
            u.setPath(path);
            //System.out.println("repathing");
        }
        return path;
    }

    private void followPath(Array<Vector2> path, Unit u) {
        if(path == null) return;
        
        Vector2 desiredVelocity = tmp.set((Vector2) path.get(0)).sub(u.getPosition());

        if (desiredVelocity.len() < 0.5f) {
            path.removeIndex(0);
            //System.out.println(path.size + " way points left");
            if (path.size == 0) {
                u.setPath(null);
                u.setState(Unit.State.IDLE);
                return;
            }
            desiredVelocity.set((Vector2) path.get(0)).sub(u.getPosition());
        }

        desiredVelocity.nor().scl(u.getSpeed()).sub(u.getVelocity());
        u.getAcceleration().add(desiredVelocity.scl(0.10f));
    }
    
    private void seek(Engine.Controls engineControls, Unit u) {
        boolean set = false;
        float[] dx = {-1.0f, 1.0f, 0.0f, 0.0f};
        float[] dy = {0.0f, 0.0f, -1.0f, 1.0f};
        float x = (int)u.getPosition().x + 0.5f;
        float y = (int)u.getPosition().y + 0.5f;
        CollisionMap map = engineControls.getCollisionMap();
        for(int i = 0; i < dx.length; i++) {
            if(map.isWalkable(x + dx[i], y + dy[i])) {
                if(!set) {
                    tmp.set(dx[i], dy[i]);
                    set = true;
                } else {
                    if(target.getPosition().dst(x + tmp.x, y + tmp.y) > target.getPosition().dst(x + dx[i], y + dy[i])) {
                        tmp.set(dx[i], dy[i]);
                    }
                }
            }
        }
        
        if(set) {
            u.getAcceleration().add(tmp.scl(0.0025f));
        }
    }

    private void separate(Engine.Controls engineControls, Unit u) {
        for (Unit otherUnit : engineControls.getEntityManager().getUnits(u.getPosition(), 5.0f)) {
            if (otherUnit != u) {
                tmp.set(u.getPosition()).sub(otherUnit.getPosition());
                float distance = tmp.len();
                if (distance < 0.8f) {
                    u.getAcceleration().add(tmp.scl(1.0f / (distance * distance * distance) * 10.0f));
                }
            }
        }
    }

    @Override
    public void dispose() {
        for(Vector2 v : visited) {
            vectorPool.free(v);
        }
    }
}
