package com.pancik.ciernypetrzlen.engine.component.entity.units;

import com.badlogic.gdx.graphics.Color;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.math.Vector3;
import com.pancik.ciernypetrzlen.engine.Animation;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.SoundData;
import com.pancik.ciernypetrzlen.engine.component.entity.Attackable;
import com.pancik.ciernypetrzlen.engine.component.entity.Fire;
import com.pancik.ciernypetrzlen.engine.component.entity.Monster;
import com.pancik.ciernypetrzlen.engine.component.entity.projectile.SkeletonBossFireball;
import com.pancik.ciernypetrzlen.engine.component.entity.projectile.Projectile;
import com.pancik.ciernypetrzlen.engine.component.particle.SkeletonDeathParticle;
import com.pancik.ciernypetrzlen.engine.component.spell.Buff;
import com.pancik.ciernypetrzlen.engine.pathfinding.BresenhamCollisionChecker;
import com.pancik.ciernypetrzlen.engine.player.YellText;

public class SkeletonBoss extends Monster {
    public static final Vector2 SIZE = new Vector2(1.25f, 1.25f);
    public static final int HEALTH = 500;
    public static final float SPEED = 0.5f / 60.0f;
    public static final int ATTACK_DAMAGE = 10;
    public static final float ATTACK_RANGE = 5.0f;
    public static final int ATTACK_STEP = 5;
    public static final int ATTACK_COOLDOWN = 120;
    
    protected static final String textureName = "unit-skeleton-mage";
    protected static TextureRegion[][] stand = Animation.cutAnimation3way(textureName, 0, 16, 16, 2);
    protected static TextureRegion[][] walk = Animation.cutAnimation3way(textureName, 32, 16, 16, 4);
    protected static TextureRegion[][] attack = Animation.cutAnimation3way(textureName, 96, 16, 16, 4);
    
    protected int failedShots = 0;
    protected int failedShotsRecoveryStep = 0;
    protected final int maxFailedShots = 2;
    protected final int failedShotsRecoveryLength = 180;

    protected int attacksCount = 0;
    
    public SkeletonBoss(Vector2 position, int team, int level, Engine.Controls engineControls) {
        super(position, SIZE, team, engineControls,
                new MonsterStats(level, HEALTH, 1.0f, SPEED, ATTACK_DAMAGE, ATTACK_RANGE, ATTACK_STEP, ATTACK_COOLDOWN));
    }

    @Override
    protected TextureRegion[][] getStandAnimation() {
        return stand;
    }

    @Override
    protected TextureRegion[][] getWalkAnimation() {
        return walk;
    }

    @Override
    protected TextureRegion[][] getAttackAnimation() {
        return attack;
    }

    @Override
    public void tick() {
        super.tick();

        if (failedShotsRecoveryStep > 0) {
            failedShotsRecoveryStep--;
        }
    }

    @Override
    public boolean isInRangeForAttack(Attackable a) {
        return super.isInRangeForAttack(a)
                && BresenhamCollisionChecker.isInLineOfSight(engineControls.getCollisionMap(), position.x, position.y, a.getPosition().x, a.getPosition().y)
                && failedShotsRecoveryStep == 0;
    }

    @Override
    public void attack(Attackable a) {
        super.attack(a);

        if (attackCooldown == 0) {
            setAttack(getAttackAnimation()[0].length * stats.attackStep, 3 * stats.attackStep, a,
                    (int)(((float)stats.attackDamage + getScalarModifier(Buff.Stat.ATTACK)) * getPercentualModifier(Buff.Stat.ATTACK)));
            attackCooldown = (int)((float)stats.attackCooldownLength * (1.0f/((1.0f + getScalarModifier(Buff.Stat.ATTACK_SPEED)) * getPercentualModifier(Buff.Stat.ATTACK_SPEED))));
        }
    }

    @Override
    protected void doAttack(int damage, Attackable a) {
        attacksCount++;
        if(attacksCount >= 10) {
            attacksCount -= 10;
            
            String[] yellTexts = {"YOU WILL DIE!", "YOU CAN'T DEFEAT ME!", "I WILL CRUSH YOU!", "AHAHAHA EAT SOME FIRE!"};
            
            Vector3 yellPosition = new Vector3();
            yellPosition.set(position.x, 0.0f, position.y - SIZE.y);
            engineControls.getPlayer().getTextEffectManager().addTextEffect(new YellText(engineControls,
                    yellPosition,
                    yellTexts[MathUtils.random(yellTexts.length - 1)],
                    Color.RED));
        }
        
        final Attackable target = a;
        final int targetDamage = damage;
        engineControls.addEntity(new SkeletonBossFireball(target.getPosition().cpy(), position.cpy(), engineControls, new Projectile.Callback() {
            @Override
            public void onProjectileArrived(Vector2 pos) {
                engineControls.addEntity(new Fire(pos.sub(0.0f, +0.25f), target, targetDamage, false, engineControls));
                failedShots = 0;
            }

            @Override
            public void onProjectileCrashed(Vector2 pos) {
                engineControls.addEntity(new Fire(pos, target, targetDamage, false, engineControls));
                
                failedShots++;
                if (failedShots > maxFailedShots) {
                    failedShots = 0;
                    failedShotsRecoveryStep = failedShotsRecoveryLength;
                }
            }
        }));
    }
        
    @Override
    public String getName() {
        return "Big Bad Skeleton " + super.getName();
    }
    
    @Override
    public void onRemove() {
        super.onRemove();
        
        SoundData.playSound("bones", 0.25f);
        engineControls.addEntity(new SkeletonDeathParticle(position, 1.25f, engineControls));
    }
}
