package com.pancik.ciernypetrzlen.engine.component.level;

import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.component.entity.Door;
import com.pancik.ciernypetrzlen.engine.component.entity.Entity;
import com.pancik.ciernypetrzlen.engine.component.entity.Monster;
import com.pancik.ciernypetrzlen.engine.component.entity.PickableItem;
import com.pancik.ciernypetrzlen.engine.component.entity.Stairs;
import com.pancik.ciernypetrzlen.engine.pathfinding.CollisionMap;
import com.pancik.ciernypetrzlen.engine.player.inventory.Item;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

public class Level {
    public static class Point {
        protected int x,y;

        public Point(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public int getX() {
            return x;
        }

        public int getY() {
            return y;
        }
    }

    private int width;
    private int height;
    private Tile[][] tileLayout;
    private Character[][] scriptLayout = null;
    private Vector2 spawnPoint;
    private List<Entity> entities = new ArrayList<Entity>();
    
    public Level(FileHandle file, Engine.Controls engineControls) {
        if(file.exists()) {
            BufferedReader input = new BufferedReader(file.reader());
            String rawInput;
            int line = 0;
            try {
                while((rawInput = input.readLine()) != null) {
                    if(line == 0) {
                        //System.out.println("1 " + rawInput);
                        String[] parsed = rawInput.split(" ");
                        width = Integer.parseInt(parsed[0]);
                        height = Integer.parseInt(parsed[1]);
                        
                        tileLayout = new Tile[width][height];
                        scriptLayout = new Character[width][height];
                    } else if(line > 0 && line <= height) {
                        //System.out.println("2 " + rawInput);
                        for(int i = 0; i < width; i++) {
                            Tile type = null;

                            switch(rawInput.charAt(i)) {
                                case '.':
                                    type = Tile.FLOOR;
                                    break;
                                case 'W':
                                    type = Tile.WALL;
                                    break;
                                case ' ':
                                    type = Tile.EMPTY;
                                    break;
                                default:
                                    throw new IllegalArgumentException("Missing character recognition for: " + rawInput.charAt(i));
                            }
                            tileLayout[i][line - 1] = type;
                        }
                    } else if(line > height + 1 && line <= (height) * 2 + 1) {
                        if(rawInput.length() == width) {
                            for(int i = 0; i < width; i++) {
                                scriptLayout[i][line - (height + 2)] = rawInput.charAt(i);
                            }
                        }
                    } else if(line > (height) * 2 + 1){
                        //System.out.println("4 " + rawInput);
                        if(rawInput.length() > 2) {
                            String[] parsed = rawInput.split(" ");
                            Point[] scriptPositions = getCharacterPosInScriptLayout(parsed[0].charAt(0));

                            for(Point scriptPos : scriptPositions) {
                                Vector2 position = new Vector2(scriptPos.x + 0.5f, scriptPos.y + 0.5f);
                                
                                if(parsed[1].equals("begin")) {
                                    spawnPoint = position;
                                } else if(parsed[1].equals("debris")) {
                                    try {
                                        DebrisType debrisType = DebrisType.valueOf(parsed[2]);
                                        entities.add(new Debris(position, debrisType, engineControls));
                                    } catch(IllegalArgumentException ex) {
                                        System.err.println("Unknown debris type of: " + parsed[2]);
                                    }
                                } else if(parsed[1].equals("door")) {
                                    String text = null;
                                    if(parsed.length > 2) {
                                        text = parsed[2].replace("^", " ");
                                    }
                                    entities.add(new Door(position, engineControls, text, false, false, null));
                                } else if(parsed[1].equals("lockedDoor")) {
                                    String text = null;
                                    if(parsed.length > 2) {
                                        text = parsed[2].replace("^", " ");
                                    }
                                    entities.add(new Door(position, engineControls, text, true, true, null));
                                } else if(parsed[1].equals("keyDoor") && parsed.length > 2) {
                                    String key = null, text = null;
                                    key = parsed[2];
                                    if(parsed.length > 3) {
                                        text = parsed[3].replace("^", " ");
                                    }
                                    entities.add(new Door(position, engineControls, text, true, false, key));
                                } else if(parsed[1].equals("stairs") && parsed.length > 2) {
                                    String level = null;
                                    if(parsed.length > 3) {
                                        level = parsed[3];
                                    }
                                    
                                    if(parsed[2].equals("up")) {
                                        entities.add(new Stairs(position, level, true, engineControls));
                                    } else if(parsed[2].equals("down")) {
                                        entities.add(new Stairs(position, level, false, engineControls));
                                    }
                                } else if(parsed[1].equals("spawn") && parsed.length > 3) {
                                    String itemClass = null;
                                    if(parsed.length > 4) {
                                        itemClass = parsed[4];
                                    }
                                    try {
                                        Class clazz;
                                        Item item = null;
                                        if(itemClass != null) {
                                            clazz = Class.forName("com.pancik.ciernypetrzlen.engine.player.inventory." + itemClass);
                                            item = (Item) clazz.newInstance();
                                        }
                                        
                                        clazz = Class.forName("com.pancik.ciernypetrzlen.engine.component.entity.units." + parsed[2]);
                                        Monster monster = (Monster)clazz
                                                .getConstructor(Vector2.class, int.class, int.class, Engine.Controls.class)
                                                .newInstance(position.add(MathUtils.random(0.6f) - 0.3f, MathUtils.random(0.6f) - 0.3f), 2, Integer.parseInt(parsed[3]), engineControls);
                                        if(item != null) {
                                            monster.addItemToLoot(item);
                                        }
                                        entities.add(monster);
                                    } catch (Exception e) {
                                        e.printStackTrace(System.err);
                                    }
                                } else if(parsed[1].equals("item") && parsed.length > 2) {
                                    try {
                                        Class clazz = Class.forName("com.pancik.ciernypetrzlen.engine.player.inventory." + parsed[2]);
                                        Item item = (Item)clazz.newInstance();
                                        entities.add(new PickableItem(position.add(MathUtils.random(0.6f) - 0.3f, MathUtils.random(0.6f) - 0.3f),
                                                item,
                                                engineControls));
                                    } catch (Exception e) {
                                        e.printStackTrace(System.err);
                                    }
                                }
                            }
                        }
                    }
                    line++;
                }
            } catch (IOException ex) {
                Logger.getLogger(Level.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
            }
        }
    }
    
    public Tile[][] getTileLayout() {
        return tileLayout;
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }
    
    public Vector2 getSpawnPoint() {
        return spawnPoint;
    }

    public List<Entity> getEntities() {
        return entities;
    }
    
    public CollisionMap buildCollisionMap() {
        boolean[][] walkableTiles = new boolean[tileLayout.length][tileLayout[0].length];
        for(int j = 0; j < tileLayout.length; j++) {
            for(int i = 0; i < tileLayout[j].length; i++) {
                walkableTiles[j][i] = (tileLayout[j][i] == Tile.FLOOR) ? true : false;
            }
        }
    
        return new CollisionMap(walkableTiles);
    }
    
    protected final Point[] getCharacterPosInScriptLayout(Character c) {
        ArrayList<Point> list = new ArrayList<Point>();
        
        for(int j = 0; j < scriptLayout.length; j++) {
            for(int i = 0; i < scriptLayout[j].length; i++) {
                if(scriptLayout[j][i].equals(c)) {
                    list.add(new Point(j, i));
                }
            }
        }
        
        Point[] out = new Point[list.size()];
        list.toArray(out);
        return out;
    }
}
