package com.pancik.ciernypetrzlen.engine.component.level;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.files.FileHandle;
import com.badlogic.gdx.graphics.Mesh;
import com.badlogic.gdx.graphics.Pixmap;
import com.badlogic.gdx.graphics.Texture;
import com.badlogic.gdx.graphics.VertexAttribute;
import com.badlogic.gdx.graphics.g2d.PixmapPacker;
import com.badlogic.gdx.graphics.g2d.TextureAtlas;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.utils.FloatArray;

public class LevelMeshes {
    private Level level;
    private Mesh[][] meshes;
    private TextureAtlas tileAtlas;
    private TextureRegion[] floorTiles;
    private TextureRegion[] wallTiles;
    private TextureRegion[] doorTilesNormal;
    private TextureRegion[] doorTilesPicklock;
    private TextureRegion[] doorTilesKey;

    public LevelMeshes(Level level) {
        this.level = level;
        build();
    }

    private void build() {
        buildTileAtlas();
        
        int width = (int)Math.ceil(level.getWidth() / 32.0);
        int height = (int)Math.ceil(level.getHeight() / 32.0);
        this.meshes = new Mesh[width][height];
        for (int z = 0; z < level.getHeight(); z += 32) {
            for (int x = 0; x < level.getWidth(); x += 32) {
                int sizeX = 32, sizeZ = 32;
                if(x + sizeX > level.getWidth()) {
                    sizeX = level.getWidth() - x;
                }
                if(z + sizeZ > level.getHeight()) {
                    sizeZ = level.getHeight() - z;
                }
                this.meshes[(x / 32)][(z / 32)] = build(level, x, z, sizeX, sizeZ);
            }
        }
    }

    private void buildTileAtlas() {
        PixmapPacker packer = new PixmapPacker(1024, 1024, Pixmap.Format.RGB888, 2, true);
        FileHandle dir = Gdx.files.internal("data/drawable/tiles");
        buildTiles(packer, "walls", dir.child("wall.png"));
        buildTiles(packer, "floors", dir.child("floor.png"));
        buildTiles(packer, "door-normal", dir.child("door-normal.png"));
        buildTiles(packer, "door-picklock", dir.child("door-picklock.png"));
        buildTiles(packer, "door-key", dir.child("door-key.png"));
        this.tileAtlas = packer.generateTextureAtlas(Texture.TextureFilter.Nearest, Texture.TextureFilter.Nearest, false);
        this.wallTiles = splitTiles(this.tileAtlas, "walls", 16, 16);
        this.floorTiles = splitTiles(this.tileAtlas, "floors", 16, 16);
        this.doorTilesNormal = splitTiles(this.tileAtlas, "door-normal", 16, 16);
        this.doorTilesPicklock = splitTiles(this.tileAtlas, "door-picklock", 16, 16);
        this.doorTilesKey = splitTiles(this.tileAtlas, "door-key", 16, 16);
    }

    private void buildTiles(PixmapPacker packer, String name, FileHandle tileFile) {
        Pixmap pixmap = new Pixmap(tileFile);
        packer.pack(name, pixmap);
        pixmap.dispose();
    }
    
    private TextureRegion[] splitTiles(TextureAtlas atlas, String name, int width, int height) {
        TextureRegion region = atlas.findRegion(name);
        TextureRegion[][] regions = region.split(width, height);
        TextureRegion[] tileRegions = new TextureRegion[regions.length * regions[0].length];
        int i = 0;
        for (int y = 0; y < regions[0].length; y++) {
            for (int x = 0; x < regions.length; x++) {
                tileRegions[(i++)] = regions[x][y];
            }
        }
        return tileRegions;
    }
    
    private Mesh build(Level level, int xS, int zS, int width, int height) {
        Tile[][] tiles = level.getTileLayout();
        FloatArray vertices = new FloatArray(level.getWidth() * level.getHeight() * 6 * 8);

        for (int z = zS; z < zS + height; z++) {
            for (int x = xS; x < xS + width; x++) {
                Tile t = Tile.FLOOR;
                if(getTile(tiles, x, z) != null) {
                    t = getTile(tiles, x ,z);
                }
                
                if (t ==  Tile.FLOOR) {
                    newQuad(vertices, x, 0.0f, z, randomTile(floorTiles));
                    if (getTile(tiles, x - 1, z) == Tile.WALL) {
                        newVerticalQuad(vertices, x, z, randomTile(wallTiles), 1.0f);
                    }
                    if (getTile(tiles, x + 1, z) == Tile.WALL) {
                        newVerticalQuad(vertices, x + 1, z, randomTile(wallTiles), -1.0f);
                    }
                    if (getTile(tiles, x, z - 1) == Tile.WALL) {
                        newHorizontalQuad(vertices, x, z, randomTile(wallTiles), 1.0f);
                    }
                    if (getTile(tiles, x, z + 1) == Tile.WALL) {
                        newHorizontalQuad(vertices, x, z + 1, randomTile(wallTiles), -1.0f);
                    }
                } else if(t == Tile.WALL) {
                    int f = 0;
                    if (getTile(tiles, x + 1, z + 1) == Tile.WALL) {
                        f++;
                    }
                    if (getTile(tiles, x - 1, z + 1) == Tile.WALL) {
                        f++;
                    }
                    if (getTile(tiles, x + 1, z - 1) == Tile.WALL) {
                        f++;
                    }
                    if (getTile(tiles, x - 1, z - 1) == Tile.WALL) {
                        f++;
                    }

                    newQuad(vertices, x, 1.0f, z, randomTile(wallTiles));
                } else if(t == Tile.EMPTY) {
                    if (getTile(tiles, x - 1, z) == Tile.WALL) {
                        newVerticalQuad(vertices, x, z, randomTile(wallTiles), 1.0f);
                    }
                    if (getTile(tiles, x + 1, z) == Tile.WALL) {
                        newVerticalQuad(vertices, x + 1, z, randomTile(wallTiles), -1.0f);
                    }
                    if (getTile(tiles, x, z - 1) == Tile.WALL) {
                        newHorizontalQuad(vertices, x, z, randomTile(wallTiles), 1.0f);
                    }
                    if (getTile(tiles, x, z + 1) == Tile.WALL) {
                        newHorizontalQuad(vertices, x, z + 1, randomTile(wallTiles), -1.0f);
                    }
                }
            }
        }

        vertices.shrink();
        Mesh mesh = new Mesh(true, vertices.size / 5, 0, new VertexAttribute[]{VertexAttribute.Position(), VertexAttribute.Normal(), VertexAttribute.TexCoords(0)});
        mesh.setVertices(vertices.items);
        return mesh;
    }

    private TextureRegion randomTile(TextureRegion[] tiles) {
        return tiles[com.badlogic.gdx.math.MathUtils.random(0, tiles.length - 1)];
    }

    private Tile getTile(Tile[][] tiles, int x, int z) {
        if (x < 0) {
            return null;
        }
        if (x >= tiles.length) {
            return null;
        }
        if (z < 0) {
            return null;
        }
        if (z >= tiles[0].length) {
            return null;
        }
        return tiles[x][z];
    }

    
    private void newQuad(FloatArray v, float x, float y, float z, TextureRegion r) {
        v.add(x); v.add(y); v.add(z); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV());
        v.add(x + 1.0f); v.add(y); v.add(z); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV());
        v.add(x + 1.0f); v.add(y); v.add(z + 1.0f); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV2());
        
        v.add(x + 1.0f); v.add(y); v.add(z + 1.0f); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV2());
        v.add(x); v.add(y); v.add(z + 1.0f); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV2());
        v.add(x); v.add(y); v.add(z); v.add(0.0f); v.add(1.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV());
    }

    private void newVerticalQuad(FloatArray v, float x, float z, TextureRegion r, float n) {
        v.add(x); v.add(0.0f); v.add(z); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV());
        v.add(x); v.add(0.0f); v.add(z + 1.0f); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV());
        v.add(x); v.add(1.0f); v.add(z + 1.0f); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV2());

        v.add(x); v.add(1.0f); v.add(z + 1.0f); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU2()); v.add(r.getV2());
        v.add(x); v.add(1.0f); v.add(z); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV2());
        v.add(x); v.add(0.0f); v.add(z); v.add(n); v.add(0.0f); v.add(0.0f); v.add(r.getU()); v.add(r.getV());
    }

    private void newHorizontalQuad(FloatArray v, float x, float z, TextureRegion r, float n) {
        v.add(x); v.add(0.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU()); v.add(r.getV());
        v.add(x + 1.0f); v.add(0.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU2()); v.add(r.getV());
        v.add(x + 1.0f); v.add(1.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU2()); v.add(r.getV2());

        v.add(x + 1.0f); v.add(1.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU2()); v.add(r.getV2());
        v.add(x); v.add(1.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU()); v.add(r.getV2());
        v.add(x); v.add(0.0f); v.add(z); v.add(0.0f); v.add(0.0f); v.add(n); v.add(r.getU()); v.add(r.getV());
    }

    public Mesh[][] getMeshes() {
        return this.meshes;
    }
    
    public Texture getTexture() {
        return this.floorTiles[0].getTexture();
    }
    
    public TextureRegion[] getNormalDoorTiles() {
        return doorTilesNormal;
    }
    
    public TextureRegion[] getPicklockDoorTiles() {
        return doorTilesPicklock;
    }
    
    public TextureRegion[] getKeyDoorTiles() {
        return doorTilesKey;
    }
    
    public TextureRegion[] getStairsTiles() {
        return wallTiles;
    }

    public void dispose() {
        for (int i = 0; i < this.meshes.length; i++) {
            for (Mesh mesh : this.meshes[i]) {
                mesh.dispose();
            }
        }
        this.tileAtlas.dispose();
    }
}
