package com.pancik.ciernypetrzlen.engine.gui;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.utils.Disposable;
import com.pancik.ciernypetrzlen.DrawableData;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.player.Player;
import com.pancik.ciernypetrzlen.util.RenderUtils;
import java.util.ArrayList;

public class GameMenuWindow extends UIWindow implements Disposable {
    private TextureRegion textureButtonUp = new TextureRegion(DrawableData.textureAtlas.findRegion("gui-menu-button"), 0, 0, 128, 32);
    private TextureRegion textureButtonDown = new TextureRegion(DrawableData.textureAtlas.findRegion("gui-menu-button"), 0, 32, 128, 32);
    private TextureRegion textureSolidWhite = DrawableData.textureAtlas.findRegion("solid-white");
    private InputHandler inputHandler;
    
    private ArrayList<Button> buttons = new ArrayList<Button>();
    private boolean mousePressed = false;
    
    public GameMenuWindow(Player player, Engine.Controls engineControls) {
        super(player, engineControls, 128, 159);
        this.inputHandler = new InputHandler();
        engineControls.getMainControls().getInputMultiplexer().addProcessor(0, inputHandler);
        
        buttons.add(new MenuButton(0, 16, "Resume", new ButtonCallback() {

            @Override
            public void onClick() {
                GameMenuWindow.this.player.hideUI();
            }
        }));
        
        buttons.add(new MenuButton(0, 53, "Inventory (E)", new ButtonCallback() {

            @Override
            public void onClick() {
                GameMenuWindow.this.player.showInventory();
            }
        }));
        
        buttons.add(new MenuButton(0, 90, "Character (W)", new ButtonCallback() {

            @Override
            public void onClick() {
                GameMenuWindow.this.player.showCharacter();
            }
        }));
        
        buttons.add(new MenuButton(0, 127, "Exit To Menu", new ButtonCallback() {

            @Override
            public void onClick() {
                GameMenuWindow.this.engineControls.goToMainMenu();
            }
        }));
    }
    
    @Override
    public void tick() {
        if(visible) {
        }
    }
    
    @Override
    public void renderUI() {
        if(visible) {
            float mouseX = Gdx.input.getX();
            float mouseY = Gdx.input.getY();
            
            DrawableData.font.setScale(getFontScale() * 2.0f);
            RenderUtils.blitText("PAUSE MENU",
                    getWindowTopLeftX() + ( 64 * sizeScale), getWindowTopLeftY(),
                    RenderUtils.TextAlligment.MIDDLE, RenderUtils.TextAlligment.BOTTOM);
            
            for(Button b : buttons) {
                if(mousePressed && b.hits(mouseX, mouseY)) {
                    b.clicked = true;
                } else {
                    b.clicked = false;
                }

                b.render();

                if(!mousePressed && b.hits(mouseX, mouseY)) {
                    b.highlight();
                }
            }
        }
    }
    
    @Override
    public void dispose() {
        engineControls.getMainControls().getInputMultiplexer().removeProcessor(inputHandler);
    }
    
    private class InputHandler extends InputAdapter {

        @Override
        public boolean touchDown(int screenX, int screenY, int pointer, int button) {
            if(visible) {
                mousePressed = true;
            }

            return false;
        }
        
        @Override
        public boolean touchUp(int screenX, int screenY, int pointer, int button) {
            if(visible) {
                if(mousePressed) {
                    for(Button b : buttons) {
                        if(b.hits(screenX, screenY)) {
                            b.callback.onClick();
                        }
                    }
                }

                mousePressed = false;
            }
            return false;
        }
    }
    
    protected interface ButtonCallback {
        public void onClick();
    }
    
    protected abstract class Button {
        public int posx;
        public int posy;
        public boolean clicked = false;
        public ButtonCallback callback;

        public Button(int posx, int posy, ButtonCallback callback) {
            this.posx = posx;
            this.posy = posy;
            this.callback = callback;
        }
        
        public abstract void render();
        public abstract void highlight();
        public abstract boolean hits(float x, float y);
    }
    
    protected class MenuButton extends Button {
        private static final int SIZE_X = 128;
        private static final int SIZE_Y = 32;
        public String text;
        
        public MenuButton(int posx, int posy, String text, ButtonCallback callback) {
            super(posx, posy, callback);
            this.posx = posx;
            this.posy = posy;
            this.text = text;
            this.callback = callback;
        }
        
        
        @Override
        public void render() {
            int newX = getWindowTopLeftX() + (posx * sizeScale);
            int newY = getWindowTopLeftY() + (posy * sizeScale);
            int sizeX = SIZE_X * sizeScale;
            int sizeY = SIZE_Y * sizeScale;

            if (!clicked) {
                RenderUtils.blit(textureButtonUp, newX, newY, sizeX, sizeY, 0.0f, false);
            } else {
                RenderUtils.blit(textureButtonDown, newX, newY, sizeX, sizeY, 0.0f, false);
            }

            DrawableData.font.setScale(getFontScale() * 2.0f);
            RenderUtils.blitText(text,
                    newX + sizeX / 2,
                    newY + sizeY / 2,
                    RenderUtils.TextAlligment.MIDDLE,
                    RenderUtils.TextAlligment.MIDDLE);
        }

        @Override
        public void highlight() {
            int newX = getWindowTopLeftX() + ((posx + 2) * sizeScale);
            int newY = getWindowTopLeftY() + ((posy + 2) * sizeScale);
            int sizeX = (SIZE_X - 4) * sizeScale;
            int sizeY = (SIZE_Y - 4) * sizeScale;
            DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 0.1f);
            DrawableData.spriteBatch.draw(textureSolidWhite, newX, newY, sizeX, sizeY);
            DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 1.0f);
        }

        @Override
        public boolean hits(float x, float y) {
            int newX = getWindowTopLeftX() + (posx * sizeScale);
            int newY = getWindowTopLeftY() + (posy * sizeScale);
            int sizeX = SIZE_X * sizeScale;
            int sizeY = SIZE_Y * sizeScale;
            return (x > newX && x < newX + sizeX && y > newY && y < newY + sizeY);
        }
    }
}
