package com.pancik.ciernypetrzlen.engine.gui;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.Input;
import com.badlogic.gdx.InputAdapter;
import com.badlogic.gdx.graphics.g2d.TextureRegion;
import com.badlogic.gdx.math.MathUtils;
import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.Disposable;
import com.pancik.ciernypetrzlen.DrawableData;
import com.pancik.ciernypetrzlen.engine.Engine;
import com.pancik.ciernypetrzlen.engine.SoundData;
import com.pancik.ciernypetrzlen.engine.component.entity.PickableItem;
import com.pancik.ciernypetrzlen.engine.player.Player;
import com.pancik.ciernypetrzlen.engine.player.StatsPack;
import com.pancik.ciernypetrzlen.engine.player.inventory.Equipment;
import com.pancik.ciernypetrzlen.engine.player.inventory.Item;
import com.pancik.ciernypetrzlen.engine.player.inventory.Usable;
import com.pancik.ciernypetrzlen.util.HorizontalAlignment;
import com.pancik.ciernypetrzlen.util.RenderUtils;
import com.pancik.ciernypetrzlen.util.VerticalAlignment;

public class InventoryWindow extends UIWindow implements Disposable {
    private TextureRegion textureInventory = DrawableData.textureAtlas.findRegion("gui-inventory");
    private TextureRegion textureSolidWhite = DrawableData.textureAtlas.findRegion("solid-white");
    private InputHandler inputHandler;
    private StatsPack statsPack;
    private Equipment.Slot[] equipmentTypes = {Equipment.Slot.SWORD, Equipment.Slot.SHIELD, Equipment.Slot.HEAD, Equipment.Slot.CHEST};
    private ItemSlot picked = null;
    private ItemSlot[] equipmentSlots;
    private ItemSlot[][] inventory;
    private ItemSlot transmuteSlot;
    
    public InventoryWindow(Player player, Engine.Controls engineControls, PersistenceData data) {
        super(player, engineControls, 128, 128);
        this.inputHandler = new InputHandler();
        engineControls.getMainControls().getInputMultiplexer().addProcessor(0, inputHandler);
        this.statsPack = player.getStatsPack();
        
        int[] posx = {11, 11, 69, 69};
        int[] posy = {6, 27, 6, 27};
        equipmentSlots = new ItemSlot[posx.length];
        for(int i = 0; i < posx.length; i++) {
            equipmentSlots[i] = new ItemSlot(posx[i], posy[i]);
            
            if(data != null && data.equipment[i] != null) {
                try {
                    Class clazz = Class.forName(data.equipment[i]);
                    equipmentSlots[i].item = (Item)clazz.newInstance();
                } catch (Exception e) {
                    System.err.println("Loading an item (" + data.equipment[i] + ") has failed.");
                    e.printStackTrace(System.err);
                }
            }
        }

        transmuteSlot = new ItemSlot(99, 6);
        if(data != null && data.transmuteSlot != null) {
            try {
                Class clazz = Class.forName(data.transmuteSlot);
                transmuteSlot.item = (Item)clazz.newInstance();
            } catch (Exception e) {
                System.err.println("Loading an item (" + data.transmuteSlot + ") has failed.");
                e.printStackTrace(System.err);
            }
        }

        int startx = 11;
        int starty = 51;
        int rows = 4;
        int columns = 6;
        int delta = 18;
        inventory = new ItemSlot[columns][rows];
        for(int x = 0; x < columns; x++) {
            for(int y = 0; y < rows; y++) {
                inventory[x][y] = new ItemSlot(startx + (x * delta), starty + (y * delta));
                
                if(data != null && data.inventory[x][y] != null) {
                    try {
                        Class clazz = Class.forName(data.inventory[x][y]);
                        inventory[x][y].item = (Item)clazz.newInstance();
                    } catch (Exception e) {
                        System.err.println("Loading an item (" + data.inventory[x][y] + ") has failed.");
                        e.printStackTrace(System.err);
                    }
                }
            }
        }
    }
    
    @Override
    public void tick() {
        if(visible) {
            statsPack.setEquipmentPointsAttack(getEquipmentStatModifier(Equipment.StatType.ATTACK));
            statsPack.setEquipmentPointsDefense(getEquipmentStatModifier(Equipment.StatType.DEFENSE));
            statsPack.setEquipmentPointsVitality(getEquipmentStatModifier(Equipment.StatType.VITALITY));
        }
        
        if(transmuteSlot.item != null && transmuteSlot.item.getTransmuteValue() > 0) {
            statsPack.changeGold(transmuteSlot.item.getTransmuteValue());
            transmuteSlot.item = null;
            
            SoundData.playSound("transmute", 0.5f);
        }
    }
    
    @Override
    public void renderUI() {
        if(visible) {
            RenderUtils.blit(textureInventory,
                    HorizontalAlignment.MIDDLE, 0,
                    VerticalAlignment.MIDDLE, 0,
                    128 * sizeScale, 128 * sizeScale,
                    0.0f,
                    false);
            
            DrawableData.font.setScale(getFontScale());
            RenderUtils.blitText(Integer.toString(statsPack.getGold()),
                    getWindowTopLeftX() + (119 * sizeScale),
                    getWindowTopLeftY() + (44 * sizeScale),
                    RenderUtils.TextAlligment.RIGHT,
                    RenderUtils.TextAlligment.MIDDLE);
            
            player.getHero().renderCharacter2D(getWindowTopLeftX() + (34 * sizeScale), getWindowTopLeftY() + (11 * sizeScale), (28 * sizeScale));
            
            float mouseX = Gdx.input.getX();
            float mouseY = Gdx.input.getY();
            
            Item showTooltip = null;
            
            for(ItemSlot s : equipmentSlots) {
                s.render();
                
                if(s.hits(mouseX, mouseY)) {
                    s.highlight();
                    showTooltip = s.item;
                }
            }

            for(int j = 0; j < inventory.length; j++) {
                for(int i = 0; i < inventory[j].length; i++) {
                    inventory[j][i].render();
                    
                    if(inventory[j][i].hits(mouseX, mouseY)) {
                        inventory[j][i].highlight();
                        showTooltip = inventory[j][i].item;
                    }
                }
            }
            
            transmuteSlot.render();
            if(transmuteSlot.hits(mouseX, mouseY)) {
                transmuteSlot.highlight();
                showTooltip = transmuteSlot.item;
            }

            if(picked != null && picked.item != null) {
                picked.item.render(mouseX - (8 * sizeScale), mouseY - (8 * sizeScale), 16 * sizeScale);
            }
            
            if(showTooltip != null) {
                showTooltip.renderTooltip(mouseX, mouseY, sizeScale);
            }
        } else {
            picked = null;
        }
    }
    
    @Override
    public void dispose() {
        engineControls.getMainControls().getInputMultiplexer().removeProcessor(inputHandler);
    }
    
    public int getItemX(Item item) {
        for (int j = 0; j < inventory.length; j++) {
            for (int i = 0; i < inventory[j].length; i++) {
                if(inventory[j][i].item == item) {
                    return j;
                }
            }
        }
        return -1;
    }
    
    public int getItemY(Item item) {
        for (int j = 0; j < inventory.length; j++) {
            for (int i = 0; i < inventory[j].length; i++) {
                if(inventory[j][i].item == item) {
                    return i;
                }
            }
        }
        return -1;
    }
    
    public Item getItem(int x, int y) {
        return inventory[x][y].item;
    }
    
    public Item getItem(String simpleClassName) {
        for (int j = 0; j < inventory.length; j++) {
            for (int i = 0; i < inventory[j].length; i++) {
                if(inventory[j][i].item != null && inventory[j][i].item.getClass().getSimpleName().equals(simpleClassName)) {
                    return inventory[j][i].item;
                }
            }
        }
        return null;
    }
    
    public boolean addItemToInventory(Item item) {
        for (int j = 0; j < inventory[0].length; j++) {
            for (int i = 0; i < inventory.length; i++) {
                if (inventory[i][j].item == null) {
                    inventory[i][j].item = item;
                    return true;
                }
            }
        }
        
        return false;
    }
    
    public void removeItem(Item item) {
        for (int j = 0; j < inventory[0].length; j++) {
            for (int i = 0; i < inventory.length; i++) {
                if (inventory[i][j].item == item) {
                    inventory[i][j].item = null;
                    break;
                }
            }
        }
    }
    
    public Usable[] getUsableItems() {
        int count = 0;
        for(ItemSlot[] row : inventory) {
            for(ItemSlot i : row) {
                if(i.item != null && i.item instanceof Usable) {
                    count++;
                }
            }
        }
        
        Usable[] out = new Usable[count];
        int index = 0;
        for(ItemSlot[] row : inventory) {
            for(ItemSlot i : row) {
                if(i.item != null && i.item instanceof Usable) {
                    out[index++] = (Usable)i.item;
                }
            }
        }
        return out;
    }
    
    public int getEquipmentStatModifier(Equipment.StatType type) {
        int modifier = 0;
        for(ItemSlot i : equipmentSlots) {
            if(i.item instanceof Equipment) {
                modifier += ((Equipment)i.item).getStatModifier(type);
            }
        }
        return modifier;
    }
    
    private class InputHandler extends InputAdapter {

        @Override
        public boolean touchDown(int screenX, int screenY, int pointer, int button) {
            if(visible) {
               if(button == Input.Buttons.LEFT) {
                    float topLeftX = getWindowTopLeftX();
                    float topLeftY = getWindowTopLeftY();
                    if(screenX >= topLeftX && screenX < topLeftX + (128 * sizeScale) && screenY >= topLeftY && screenY < topLeftY + (128 * sizeScale)) {
                        if(picked == null) {
                            for (ItemSlot s : equipmentSlots) {
                                if (s.hits(screenX, screenY) && s.item != null) {
                                    picked = s;
                                    return true;
                                }
                            }

                            for (int j = 0; j < inventory.length; j++) {
                                for (int i = 0; i < inventory[j].length; i++) {
                                    if (inventory[j][i].hits(screenX, screenY) && inventory[j][i].item != null) {
                                        picked = inventory[j][i];
                                        return true;
                                    }
                                }
                            }

                            if(transmuteSlot.hits(screenX, screenY) && transmuteSlot.item != null) {
                                picked = transmuteSlot;
                                return true;
                            }
                        } else {
                            ItemSlot found = null;

                            for (int i = 0; i < equipmentSlots.length; i++) {
                                if (equipmentSlots[i].hits(screenX, screenY) && picked != null && picked.item instanceof Equipment && ((Equipment)picked.item).getSlot() == equipmentTypes[i]) {
                                    found = equipmentSlots[i];
                                }
                            }

                            for (int j = 0; j < inventory.length; j++) {
                                for (int i = 0; i < inventory[j].length; i++) {
                                    if (inventory[j][i].hits(screenX, screenY)) {
                                        found = inventory[j][i];
                                    }
                                }
                            }

                            if(transmuteSlot.hits(screenX, screenY)) {
                                found = transmuteSlot;
                            }

                            if(found != null) {
                                if (picked == found) {
                                    picked = null;
                                    return true;
                                }
                                Item item = picked.item;
                                picked.item = found.item;
                                found.item = item;
                                if (picked.item == null) {
                                    picked = null;
                                }
                                return true;
                            }
                        }
                    } else {
                        if(picked != null && picked.item != null) {
                            Vector2 pos = player.getHero().getPosition().cpy();
                            pos.set((int)pos.x, (int)pos.y).add(MathUtils.random(1.0f), MathUtils.random(1.0f));
                            engineControls.addEntity(new PickableItem(pos, picked.item, engineControls));
                            picked.item = null;
                        }
                    }
                } else if(button == Input.Buttons.RIGHT) {
                    for(int j = 0; j < inventory.length; j++) {
                        for(int i = 0; i < inventory[j].length; i++) {
                            if(inventory[j][i].hits(Gdx.input.getX(), Gdx.input.getY()) && inventory[j][i].item != null && inventory[j][i].item instanceof Usable) {
                                ((Usable)inventory[j][i].item).use(player.getHero());
                                inventory[j][i].item = null;
                            }
                        }
                    }
                }
            }

            return false;
        }

        @Override
        public boolean keyDown(int keycode) {
            if(keycode == Input.Keys.T) {
                float mouseX = Gdx.input.getX();
                float mouseY = Gdx.input.getY();
                
                for(int j = 0; j < inventory.length; j++) {
                    for(int i = 0; i < inventory[j].length; i++) {
                        if(inventory[j][i].item != null && inventory[j][i].hits(mouseX, mouseY)) {
                            Item swap = transmuteSlot.item;
                            transmuteSlot.item = inventory[j][i].item;
                            inventory[j][i].item = swap;
                            return true;
                        }
                    }
                }
            }
            
            return false;
        }
    }
    
    protected class ItemSlot {
        public int posx;
        public int posy;
        public Item item = null;

        public ItemSlot(int posx, int posy) {
            this.posx = posx;
            this.posy = posy;
        }
        
        public void render() {
            if(item != null && picked != this) {
                int newX = getWindowTopLeftX() + (posx * sizeScale);
                int newY = getWindowTopLeftY() + (posy * sizeScale);
                int size = 16 * sizeScale;
                item.render(newX, newY, size);
            }
        }
        
        public void highlight() {
            int newX = getWindowTopLeftX() + (posx * sizeScale);
            int newY = getWindowTopLeftY() + (posy * sizeScale);
            int size = 16 * sizeScale;
            DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 0.3f);
            DrawableData.spriteBatch.draw(textureSolidWhite, newX, newY, size, size);
            DrawableData.spriteBatch.setColor(1.0f, 1.0f, 1.0f, 1.0f);
        }
        
        public boolean hits(float x, float y) {
            int newX = getWindowTopLeftX() + (posx * sizeScale);
            int newY = getWindowTopLeftY() + (posy * sizeScale);
            int size = 16 * sizeScale;
            return (x > newX && x < newX + size && y > newY && y < newY + size);
        }
    }
     
    public static class PersistenceData {
        public String[] equipment;
        public String[][] inventory;
        public String transmuteSlot;
    }
    
    public PersistenceData getPersistenceData() {
        PersistenceData out = new PersistenceData();
        
        out.equipment = new String[equipmentSlots.length];
        for(int i = 0; i < equipmentSlots.length; i++) {
            if(equipmentSlots[i].item != null) {
                out.equipment[i] = equipmentSlots[i].item.getClass().getName();
            } else {
                out.equipment[i] = null;
            }
        }
        
        out.inventory = new String[inventory.length][inventory[0].length];
        for(int j = 0; j < inventory.length; j++) {
            for(int i = 0; i < inventory[j].length; i++) {
                if(inventory[j][i].item != null) {
                    out.inventory[j][i] = inventory[j][i].item.getClass().getName();
                } else {
                    out.inventory[j][i] = null;
                }
            }
        }
        
        if(transmuteSlot.item != null) {
            out.transmuteSlot = transmuteSlot.item.getClass().getName();
        } else {
            out.transmuteSlot = null;
        }
        
        return out;
    }
}
