package com.pancik.ciernypetrzlen.engine.pathfinding;

import com.badlogic.gdx.math.Vector2;
import com.badlogic.gdx.utils.Array;
import com.badlogic.gdx.utils.ObjectMap;
import com.badlogic.gdx.utils.Pool;
import java.util.Iterator;

public class AStarPathFinder {
    static final Pool<Node> nodePool = new Pool() {
        protected AStarPathFinder.Node newObject() {
            return new AStarPathFinder.Node();
        }

        @Override
        public AStarPathFinder.Node obtain() {
            AStarPathFinder.Node node = (AStarPathFinder.Node) super.obtain();
            node.cell = null;
            node.parent = null;
            node.gScore = (node.hScore = node.fScore = 0.0f);
            return node;
        }
    };
    static ObjectMap<Cell, Node> nodes = new ObjectMap();
    static final ObjectMap<Cell, Cell> openSet = new ObjectMap();
    static final ObjectMap<Cell, Cell> closedSet = new ObjectMap();

    public static synchronized Array<Vector2> findPath(Cell start, Cell end, DistanceEstimator distEstimator, DistanceEstimator heuristicEstimator) {
        Node startNode = (Node) nodePool.obtain();
        startNode.cell = start;
        startNode.gScore = 0.0f;
        startNode.hScore = heuristicEstimator.getDistance(start, end);
        startNode.fScore = startNode.hScore;

        openSet.put(start, start);
        nodes.put(start, getNode(start));
        Iterator localIterator;
        while(openSet.size > 0) {
            Node x = getLowestFScoreCell();
            if (x.cell.equals(end)) {
                Array path = reconstructPath(end);
                cleanUp();
                return path;
            }

            openSet.remove(x.cell);
            closedSet.put(x.cell, x.cell);

            localIterator = x.cell.getNeighbours().iterator();
            while(localIterator.hasNext()) {
                Cell cell = (Cell) localIterator.next();
                Node y = getNode(cell);
                if (!closedSet.containsKey(y.cell)) {
                    float tentativeGScore = x.gScore + distEstimator.getDistance(x.cell, y.cell);
                    boolean tentativeIsBetter;

                    if (!openSet.containsKey(y.cell)) {
                        openSet.put(y.cell, y.cell);
                        tentativeIsBetter = true;
                    } else if (tentativeGScore < y.gScore) {
                        tentativeIsBetter = true;
                    } else {
                        tentativeIsBetter = false;
                    }

                    if (tentativeIsBetter) {
                        y.parent = x.cell;
                        y.gScore = tentativeGScore;
                        y.hScore = heuristicEstimator.getDistance(y.cell, end);
                        y.fScore = (y.gScore + y.hScore);
                    }
                }
            }
        }

        cleanUp();
        return null;
    }

    private static void cleanUp() {
        for (Node node : nodes.values()) {
            nodePool.free(node);
        }
        nodes.clear();
        openSet.clear();
        closedSet.clear();
    }

    private static Node getNode(Cell cell) {
        Node node = (Node) nodes.get(cell);
        if (node == null) {
            node = (Node) nodePool.obtain();
            node.cell = cell;
            nodes.put(cell, node);
        }
        return node;
    }

    private static Node getLowestFScoreCell() {
        float best = 123456789.0f;
        Node bestNode = null;
        for (Cell cell : openSet.keys()) {
            Node node = (Node) nodes.get(cell);
            if (node.fScore < best) {
                best = node.fScore;
                bestNode = node;
            }
        }
        return bestNode;
    }

    private static Array<Vector2> reconstructPath(Cell end) {
        Array path = new Array();
        Node node = (Node) nodes.get(end);

        while (node.parent != null) {
            path.add(node.cell.getPosition());
            node = (Node) nodes.get(node.parent);
        }
        path.reverse();
        return path.size == 0 ? null : path;
    }

    static class Node {
        Cell cell;
        Cell parent;
        float gScore;
        float fScore;
        float hScore;

        public void set(Cell cell, Cell parent, float gCost, float fCost) {
            this.parent = parent;
            this.gScore = gCost;
            this.fScore = fCost;
        }
    }
}